@@ -1,5 +1,4 @@
-#ifndef MATHS_H
-#define MATHS_H
+#pragma once
 
 #define _USE_MATH_DEFINES
 
@@ -13,6 +12,17 @@
 // [[Rcpp::depends(RcppEigen)]]
 
 namespace glmmr {
+
+template<class T>
+T randomGaussian(T generator,
+                 VectorXd& res)
+{
+  for (size_t i = 0; i < res.size(); ++i)
+    res(i) = generator();
+  // Note the generator is returned back
+  return  generator;
+}
+
 namespace maths {
 
 inline double gaussian_cdf(double value)
@@ -53,143 +63,163 @@ inline Eigen::VectorXd exp_vec(const Eigen::VectorXd& x,
   return z;
 }
 
-inline Eigen::VectorXd mod_inv_func(Eigen::VectorXd mu,
-                                    std::string link)
+inline VectorXd mod_inv_func(const VectorXd& muin,
+                                    LinkDistribution link)
 {
-  const static std::unordered_map<std::string, int> string_to_case{
-    {"logit",1},
-    {"log",2},
-    {"probit",3},
-    {"identity",4},
-    {"inverse",5}
-  };
-  switch (string_to_case.at(link)) {
-  case 1:
+  using enum LinkDistribution;
+  VectorXd mu(muin);
+  switch (link) {
+  case logit:
     mu = exp_vec(mu, true);
     break;
-  case 2:
+  case loglink:
     mu = exp_vec(mu);
     break;
-  case 3:
+  case probit:
     mu = gaussian_cdf_vec(mu);
     break;
-  case 4:
+  case identity:
     break;
-  case 5:
+  case inverse:
     mu = mu.array().inverse().matrix();
     break;
   }
   return mu;
 }
 
+
+
 inline Eigen::VectorXd dhdmu(const Eigen::VectorXd& xb,
                              const glmmr::Family& family) {
-  
+  using enum FamilyDistribution;
+  using enum LinkDistribution;
   Eigen::VectorXd wdiag(xb.size());
   Eigen::ArrayXd p(xb.size());
   
-  switch (family.flink) {
-  case 1:
-    wdiag = exp_vec(-1.0 * xb);
-    break;
-  case 2:
-    wdiag = exp_vec(xb);
-    break;
-  case 3: case 13:
-    p = mod_inv_func(xb, "logit");
-    for(int i =0; i< xb.size(); i++){
-      wdiag(i) = 1/(p(i)*(1.0 - p(i)));
-    }
-    break;
-  case 4: case 14:
-    p = mod_inv_func(xb, "log");
-    for(int i =0; i< xb.size(); i++){
-      wdiag(i) = (1.0 - p(i))/p(i);
-    }
-    break;
-  case 5: case 15:
-    p = mod_inv_func(xb, "identity");
-    wdiag = (p * (1 - p)).matrix();
-    break;
-  case 6: case 16:
-    {
-      p = mod_inv_func(xb, "probit");
-      Eigen::ArrayXd pinv = gaussian_pdf_vec(xb);
-      for(int i =0; i< xb.size(); i++){
-        wdiag(i) = (p(i) * (1-p(i)))/pinv(i);
-      }
+  switch(family.family){
+    case poisson:
+      {
+        switch(family.link){
+          case identity:
+            wdiag = exp_vec(xb);
+            break;
+          default:
+            wdiag = exp_vec(-1.0 * xb);
+            break;
+          }
       break;
-    }
-  case 7:
-    for(int i =0; i< xb.size(); i++){
-      wdiag(i) = 1.0;
-    }
-    break;
-  case 8:
-    for(int i =0; i< xb.size(); i++){
-      wdiag(i) = 1/exp(xb(i));
-    }
-    break;
-  case 9:
-    for(int i =0; i< xb.size(); i++){
-      wdiag(i) = 1.0;
-    }
-    break;
-  case 10:
-    for(int i =0; i< xb.size(); i++){
-      wdiag(i) = 1/(xb(i)*xb(i));
-    }
-    break;
-  case 11:
-    for(int i =0; i< xb.size(); i++){
-      wdiag(i) = (xb(i)*xb(i));
-    }
-    break;
-  case 12:
-    p = mod_inv_func(xb, "logit");
-    for(int i =0; i< xb.size(); i++){
-      wdiag(i) = 1/(p(i)*(1.0 - p(i)));
-    }
-    break;
+      }
+    case bernoulli: case binomial:
+      {
+        switch(family.link){
+          case loglink:
+            p = mod_inv_func(xb, family.link);
+            for(int i =0; i< xb.size(); i++){
+              wdiag(i) = (1.0 - p(i))/p(i);
+            }
+            break;
+          case identity:
+            p = mod_inv_func(xb, family.link);
+            wdiag = (p * (1 - p)).matrix();
+            break;
+          case probit:
+          {
+            p = mod_inv_func(xb, family.link);
+            Eigen::ArrayXd pinv = gaussian_pdf_vec(xb);
+            for(int i =0; i< xb.size(); i++){
+              wdiag(i) = (p(i) * (1-p(i)))/pinv(i);
+            }
+            break;
+          }
+          default:
+            p = mod_inv_func(xb, family.link);
+            for(int i =0; i< xb.size(); i++){
+              wdiag(i) = 1/(p(i)*(1.0 - p(i)));
+            }
+            break;
+        }
+      break;
+      }
+    case gaussian:
+      {
+        switch(family.link){
+          case loglink:
+            for(int i =0; i< xb.size(); i++){
+              wdiag(i) = 1/exp(xb(i));
+            }
+            break;
+          default:
+            //identity
+            for(int i =0; i< xb.size(); i++){
+              wdiag(i) = 1.0;
+            }
+            break;
+        }
+      break;
+      }
+    case gamma:
+      {
+        switch(family.link){
+          case inverse:
+            for(int i =0; i< xb.size(); i++){
+              wdiag(i) = 1/(xb(i)*xb(i));
+            }
+            break;
+          case identity:
+            for(int i =0; i< xb.size(); i++){
+              wdiag(i) = (xb(i)*xb(i));
+            }
+            break;
+          default:
+            //log
+            for(int i =0; i< xb.size(); i++){
+              wdiag(i) = 1.0;
+            }
+            break;
+        }
+      break;
+      }
+    case beta:
+      {
+        //only logit currently
+        p = mod_inv_func(xb, family.link);
+        for(int i =0; i< xb.size(); i++){
+          wdiag(i) = 1/(p(i)*(1.0 - p(i)));
+        }
+        break;
+      }
   }
   return wdiag;
 }
 
 inline Eigen::VectorXd detadmu(const Eigen::VectorXd& xb,
-                               std::string link) {
-  
+                               const LinkDistribution link) {
+  using enum LinkDistribution;
   Eigen::VectorXd wdiag(xb.size());
   Eigen::VectorXd p(xb.size());
-  const static std::unordered_map<std::string, int> string_to_case{
-    {"log",1},
-    {"identity",2},
-    {"logit",3},
-    {"probit",4},
-    {"inverse",5}
-  };
   
-  switch (string_to_case.at(link)) {
-  case 1:
+  switch (link) {
+  case loglink:
     wdiag = glmmr::maths::exp_vec(-1.0 * xb);
     break;
-  case 2:
+  case identity:
     for(int i =0; i< xb.size(); i++){
       wdiag(i) = 1.0;
     }
     break;
-  case 3:
-    p = glmmr::maths::mod_inv_func(xb, "logit");
+  case logit:
+    p = glmmr::maths::mod_inv_func(xb, link);
     for(int i =0; i< xb.size(); i++){
       wdiag(i) = 1/(p(i)*(1.0 - p(i)));
     }
     break;
-  case 4:
+  case probit:
     {
       Eigen::ArrayXd pinv = gaussian_pdf_vec(xb);
       wdiag = (pinv.inverse()).matrix();
       break;
     }
-  case 5:
+  case inverse:
     for(int i =0; i< xb.size(); i++){
       wdiag(i) = -1.0 * xb(i) * xb(i);
     }
@@ -203,14 +233,15 @@ inline Eigen::VectorXd detadmu(const Eigen::VectorXd& xb,
 inline Eigen::VectorXd attenuted_xb(const Eigen::VectorXd& xb,
                                     const Eigen::MatrixXd& Z,
                                     const Eigen::MatrixXd& D,
-                                    const std::string& link){
+                                    const LinkDistribution link){
+  using enum LinkDistribution;
   Eigen::ArrayXd xbnew(xb.array());
   int n = xb.size();
-  if(link=="log"){
+  if(link==loglink){
     for(int i=0; i<n; i++){
       xbnew(i) += (Z.row(i)*D*Z.row(i).transpose())(0)/2;
     }
-  } else if(link=="probit"){
+  } else if(link==probit){
     Eigen::ArrayXd zprod(n);
     Eigen::MatrixXd Dzz(D.rows(),D.cols());
     Eigen::PartialPivLU<Eigen::MatrixXd> pluDzz;
@@ -221,7 +252,7 @@ inline Eigen::VectorXd attenuted_xb(const Eigen::VectorXd& xb,
     }
     zprod = zprod.inverse().sqrt();
     xbnew *= zprod;
-  } else if(link=="logit"){
+  } else if(link==logit){
     double c = 0.5880842;
     Eigen::ArrayXd zprod(n);
     Eigen::MatrixXd Dzz(D.rows(),D.cols());
@@ -238,32 +269,27 @@ inline Eigen::VectorXd attenuted_xb(const Eigen::VectorXd& xb,
   return xbnew.matrix();
 }
 
+// is this function used anywhere? mark for deletion?
 inline Eigen::VectorXd marginal_var(const Eigen::VectorXd& mu,
-                                    const std::string& family,
+                                    const FamilyDistribution family,
                                     double var_par = 1.0){
+  using enum FamilyDistribution;
   Eigen::ArrayXd wdiag(mu.size());
-  const static std::unordered_map<std::string, int> string_to_case{
-    {"gaussian",1},
-    {"binomial",2},
-    {"poisson",3},
-    {"Gamma",4},
-    {"beta",5}
-  };
   
-  switch (string_to_case.at(family)) {
-  case 1:
+  switch (family) {
+  case gaussian:
     wdiag.setConstant(var_par);
     break;
-  case 2:
+  case bernoulli: case binomial:
     wdiag = mu.array()*(1-mu.array());
     break;
-  case 3:
+  case poisson:
     wdiag = mu.array();
     break;
-  case 4:
+  case gamma:
     wdiag = mu.array().square();
     break;
-  case 5:
+  case beta:
     wdiag = mu.array()*(1-mu.array())/(var_par+1);
     break;
   }
@@ -282,110 +308,154 @@ inline double log_factorial_approx(double n){
   return ans;
 }
 
-inline double log_likelihood(double y,
-                             double mu,
-                             double var_par,
-                             int flink) {
+inline double log_likelihood(const double y,
+                             const double mu,
+                             const double var_par,
+                             const FamilyDistribution family,
+                             const LinkDistribution link) {
+  using enum FamilyDistribution;
+  using enum LinkDistribution;
   double logl = 0;
-  
-  switch (flink){
-  case 1:
+  switch(family){
+  case poisson:
   {
-    double lf1 = glmmr::maths::log_factorial_approx(y);
-    logl = y * mu - exp(mu) - lf1;
-    break;
-  }
-  case 2:
+    switch(link){
+  case identity:
   {
     double lf1 = log_factorial_approx(y);
     logl = y*log(mu) - mu-lf1;
     break;
   }
-  case 3:
-    if(y==1){
-      logl = log(1/(1+exp(-1.0*mu)));
-    } else {
-      logl = log(1 - 1/(1+exp(-1.0*mu)));
+  default:
+    {
+      double lf1 = glmmr::maths::log_factorial_approx(y);
+      logl = y * mu - exp(mu) - lf1;
+      break;
     }
+  }
     break;
-  case 4:
+  }
+  case bernoulli:
+  {
+    switch(link){
+  case loglink:
     if(y==1){
       logl = mu;
     } else {
       logl = log(1 - exp(mu));
     }
     break;
-  case 5:
+  case identity:
     if(y==1){
       logl = log(mu);
     } else {
       logl = log(1 - mu);
     }
     break;
-  case 6:
-    if(y==1){
-      logl = (double)R::pnorm(mu,0,1,true,true);
-    } else {
-      logl = log(1 - (double)R::pnorm(mu,0,1,true,false));
+  case probit:
+  {
+    boost::math::normal norm(0,1);
+    if (y == 1) {
+      logl = (double)cdf(norm,mu);
+    }
+    else {
+      logl = log(1 - (double)cdf(norm, mu));
     }
     break;
-  case 7:
-    logl = -0.5*log(var_par) -0.5*log(2*3.141593) -
-      0.5*(y - mu)*(y - mu)/var_par;
+  }
+  default:
+    //logit
+    if(y==1){
+      logl = log(1/(1+exp(-1.0*mu)));
+    } else {
+      logl = log(1 - 1/(1+exp(-1.0*mu)));
+    }
     break;
-  case 8:
+  }
+    break;
+  }
+  case binomial:
+  {
+    switch(link){
+  case loglink:
+  {
+    double lfk = glmmr::maths::log_factorial_approx(y);
+    double lfn = glmmr::maths::log_factorial_approx(var_par);
+    double lfnk = glmmr::maths::log_factorial_approx(var_par - y);
+    logl = lfn - lfk - lfnk + y*mu + (var_par - y)*log(1 - exp(mu));
+    break;
+  }
+  case identity:
+  {
+    double lfk = glmmr::maths::log_factorial_approx(y);
+    double lfn = glmmr::maths::log_factorial_approx(var_par);
+    double lfnk = glmmr::maths::log_factorial_approx(var_par - y);
+    logl = lfn - lfk - lfnk + y*log(mu) + (var_par - y)*log(1 - mu);
+    break;
+  }
+  case probit:
+  {
+    boost::math::normal norm(0, 1);
+    double lfk = glmmr::maths::log_factorial_approx(y);
+    double lfn = glmmr::maths::log_factorial_approx(var_par);
+    double lfnk = glmmr::maths::log_factorial_approx(var_par - y);
+    logl = lfn - lfk - lfnk + y*((double)cdf(norm,mu)) + (var_par - y)*log(1 - (double)cdf(norm,mu));
+    break;
+  }
+  default:
+  {
+    double lfk = glmmr::maths::log_factorial_approx(y);
+    double lfn = glmmr::maths::log_factorial_approx(var_par);
+    double lfnk = glmmr::maths::log_factorial_approx(var_par - y);
+    logl = lfn - lfk - lfnk + y*log(1/(1+exp(-1.0*mu))) + (var_par - y)*log(1 - 1/(1+exp(-1.0*mu)));
+    break;
+  }
+  }
+    break;
+  }
+  case gaussian:
+  {
+    switch(link){
+  case loglink:
     logl = -0.5*log(var_par) -0.5*log(2*3.141593) -
       0.5*(log(y) - mu)*(log(y) - mu)/var_par;
     break;
-  case 9:
-      {
-        double ymu = var_par*y/exp(mu);
-        logl = log(1/(tgamma(var_par)*y)) + var_par*log(ymu) - ymu;
-        break;
-      }
-  case 10:
-      {
-        double ymu = var_par*y*mu;
-        logl = log(1/(tgamma(var_par)*y)) + var_par*log(ymu) - ymu;
-        break;
-      }
-  case 11:
+  default:
+    //identity
+    logl = -0.5*log(var_par) -0.5*log(2*3.141593) -
+      0.5*(y - mu)*(y - mu)/var_par;
+    break;
+  }
+    break;
+  }
+  case gamma:
+  {
+    switch(link){
+  case inverse:
+  {
+    double ymu = var_par*y*mu;
+    logl = log(1/(tgamma(var_par)*y)) + var_par*log(ymu) - ymu;
+    break;
+  }
+  case identity:
     logl = log(1/(tgamma(var_par)*y)) + var_par*log(var_par*y/mu) - var_par*y/mu;
     break;
-  case 12:
+  default:
+    //log
+    {
+      double ymu = var_par*y/exp(mu);
+      logl = log(1/(tgamma(var_par)*y)) + var_par*log(ymu) - ymu;
+      break;
+    }
+  }
+    break;
+  }
+  case beta:
+  {
+    //only logit currently
     logl = (mu*var_par - 1)*log(y) + ((1-mu)*var_par - 1)*log(1-y) - lgamma(mu*var_par) - lgamma((1-mu)*var_par) + lgamma(var_par);
-  case 13:
-    {
-      double lfk = glmmr::maths::log_factorial_approx(y);
-      double lfn = glmmr::maths::log_factorial_approx(var_par);
-      double lfnk = glmmr::maths::log_factorial_approx(var_par - y);
-      logl = lfn - lfk - lfnk + y*log(1/(1+exp(-1.0*mu))) + (var_par - y)*log(1 - 1/(1+exp(-1.0*mu)));
-      break;
-    }
-  case 14:
-    {
-      double lfk = glmmr::maths::log_factorial_approx(y);
-      double lfn = glmmr::maths::log_factorial_approx(var_par);
-      double lfnk = glmmr::maths::log_factorial_approx(var_par - y);
-      logl = lfn - lfk - lfnk + y*mu + (var_par - y)*log(1 - exp(mu));
-      break;
-    }
-  case 15:
-    {
-      double lfk = glmmr::maths::log_factorial_approx(y);
-      double lfn = glmmr::maths::log_factorial_approx(var_par);
-      double lfnk = glmmr::maths::log_factorial_approx(var_par - y);
-      logl = lfn - lfk - lfnk + y*log(mu) + (var_par - y)*log(1 - mu);
-      break;
-    }
-  case 16:
-    {
-      double lfk = glmmr::maths::log_factorial_approx(y);
-      double lfn = glmmr::maths::log_factorial_approx(var_par);
-      double lfnk = glmmr::maths::log_factorial_approx(var_par - y);
-      logl = lfn - lfk - lfnk + y*((double)R::pnorm(mu,0,1,true,true)) + (var_par - y)*log(1 - (double)R::pnorm(mu,0,1,true,false));
-      break;
-    }
+    break;
+  }
   }
   return logl;
 }
@@ -426,6 +496,3 @@ inline std::vector<int> linseq(int start, int end) {
 }
 }
 }
-
-
-#endif
