@@ -2,6 +2,8 @@
 
 #include "general.h"
 #include "covariance.hpp"
+#include "hsgpcovariance.hpp"
+#include "nngpcovariance.hpp"
 #include "linearpredictor.hpp"
 #include "family.hpp"
 #include "modelextradata.hpp"
@@ -23,28 +25,29 @@ public:
   glmmr::calculator calc;
   glmmr::calculator vcalc;
   bool weighted = false;
-  ModelBits(const std::string& formula_,
-            const ArrayXXd& data_,
-            const strvec& colnames_,
-            std::string family_, 
-            std::string link_) : 
-    formula(formula_), 
-    covariance(formula,data_,colnames_),
-    linear_predictor(formula,data_,colnames_),
-    data(data_.rows()),
-    family(family_,link_) { setup_calculator(); };
-  
-  int n(){return linear_predictor.n();};
-  ArrayXd xb(){return linear_predictor.xb() + data.offset;};
+  ModelBits(const std::string& formula_,const ArrayXXd& data_,const strvec& colnames_,std::string family_,std::string link_);
+  virtual int n() const {return linear_predictor.n();};
+  virtual ArrayXd xb() {return linear_predictor.xb().array() + data.offset.array();};
   virtual void make_covariance_sparse();
   virtual void make_covariance_dense();
-  
 private:
   void setup_calculator();
 };
 
 }
 
+template<typename cov, typename linpred>
+inline glmmr::ModelBits<cov, linpred>::ModelBits(const std::string& formula_,
+          const ArrayXXd& data_,
+          const strvec& colnames_,
+          std::string family_, 
+          std::string link_) : 
+  formula(formula_), 
+  covariance(formula,data_,colnames_),
+  linear_predictor(formula,data_,colnames_),
+  data(data_.rows()),
+  family(family_,link_) { setup_calculator(); };
+
 template<typename cov, typename linpred>
 inline void glmmr::ModelBits<cov, linpred>::setup_calculator(){
   dblvec yvec(n(),0.0);
@@ -63,6 +66,12 @@ inline void glmmr::ModelBits<cov, linpred>::setup_calculator(){
   vcalc.variance = data.variance;
 }
 
+template<>
+inline void glmmr::ModelBits<glmmr::hsgpCovariance, glmmr::LinearPredictor>::setup_calculator(){
+  int i = 0;
+  (void)i;
+}
+
 template<typename cov, typename linpred>
 void glmmr::ModelBits<cov, linpred>::make_covariance_sparse(){
   covariance.set_sparse(true);
@@ -72,3 +81,11 @@ template<typename cov, typename linpred>
 void glmmr::ModelBits<cov, linpred>::make_covariance_dense(){
   covariance.set_sparse(false);
 }
+
+typedef glmmr::Covariance covariance;
+typedef glmmr::nngpCovariance nngp;
+typedef glmmr::hsgpCovariance hsgp;
+typedef glmmr::LinearPredictor xb;
+typedef glmmr::ModelBits<covariance, xb> bits;
+typedef glmmr::ModelBits<nngp, xb> bits_nngp;
+typedef glmmr::ModelBits<hsgp, xb> bits_hsgp;
